use io;

/* Read instance data. */
function input() {
    local usage = "Usage: localsolver stochastic_jobshop.lsp inFileName=instanceFile "
            + "[outFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    inFile = io.openRead(inFileName);
    inFile.readln();
    nbJobs = inFile.readInt();
    nbMachines = inFile.readInt();
    nbScenarios = inFile.readInt();
    inFile.readln();
    
    // Processing times for each job on each machine (given in the processing order)
    for [s in 0...nbScenarios][j in 0...nbJobs][m in 0...nbMachines] {
        processingTimesInProcessingOrderPerScenario[s][j][m] = inFile.readInt();
    }

    inFile.readln();
    for [j in 0...nbJobs][k in 0...nbMachines] {
        local m = inFile.readInt()-1;

        // Processing order of machines for each job
        machineOrder[j][k] = m;

        for [s in 0...nbScenarios] {
            // Reorder processing times: processingTime[s][j][m] is the processing time of the
            // task of job j that is processed on machine m in the scenario s
            processingTimePerScenario[s][j][m] = processingTimesInProcessingOrderPerScenario[s][j][k];
        }
    }
    inFile.close();

    // Trivial upper bounds for the start times of the tasks
    maxStart = max[s in 0...nbScenarios](sum[j in 0...nbJobs][m in 0...nbMachines](processingTimePerScenario[s][j][m]));
}


/* Declare the optimization model */
function model() {
    // Interval decisions: time range of each task
    // tasks[s][j][m] is the interval of time of the task of job j which is processed
    // on machine m in the scenario s
    tasks[s in 0...nbScenarios][j in 0...nbJobs][m in 0...nbMachines] <- interval(0, maxStart);

    // Task duration constraints
    for [s in 0...nbScenarios][j in 0...nbJobs][m in 0...nbMachines]
        constraint length(tasks[s][j][m]) == processingTimePerScenario[s][j][m];

    // Precedence constraints between the tasks of a job
    for [s in 0...nbScenarios][j in 0...nbJobs][k in 0...nbMachines-1]
        constraint tasks[s][j][machineOrder[j][k]] < tasks[s][j][machineOrder[j][k + 1]];

    // Sequence of tasks on each machine
    jobsOrder[m in 0...nbMachines] <- list(nbJobs);

    for [m in 0...nbMachines] {
        // Each job has a task scheduled on each machine
        constraint count(jobsOrder[m]) == nbJobs;

        // Disjunctive resource constraints between the tasks on a machine
        for [s in 0...nbScenarios] {
            constraint and(0...nbJobs-1,
                    i => tasks[s][jobsOrder[m][i]][m] < tasks[s][jobsOrder[m][i + 1]][m]);
        }
    }

    // Minimize the maximum makespan: end of the last task of the last job
    // over all scenarios
    makespans[s in 0...nbScenarios] <- max[j in 0...nbJobs](end(tasks[s][j][machineOrder[j][nbMachines - 1]]));
    maxMakespan <- max[s in 0...nbScenarios](makespans[s]);
    minimize maxMakespan;
}

/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

/* Write the solution in a file with the following format:
 *  - for each machine, the job sequence */
function output() {
    if (outFileName != nil) {
        outFile = io.openWrite(outFileName);
        println("Solution written in file ", outFileName);
        for [m in 0...nbMachines]
            outFile.println[j in 0...nbJobs](jobsOrder[m].value[j], " ");
    }
}
